from plone.autoform.form import AutoExtensibleForm
from plone.base.interfaces import ISelectableConstrainTypes
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from z3c.form import form
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope.i18nmessageid import MessageFactory
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import invariant
from zope.interface.exceptions import Invalid
from zope.schema import Choice
from zope.schema import List
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


# XXX
# acquire locallyAllowedTypes from parent (default)
ACQUIRE = -1

# use default behavior of PortalFolder which uses the FTI information
DISABLED = 0

# allow types from locallyAllowedTypes only
ENABLED = 1


def ST(key, title):
    return SimpleTerm(value=key, title=title)


# reuse the translations that we had in atcontenttypes
_ = MessageFactory("plone")

possible_constrain_types = SimpleVocabulary(
    [
        ST(
            ACQUIRE,
            _("constraintypes_acquire_label", default="Use parent folder settings"),
        ),
        ST(DISABLED, _("constraintypes_disable_label", default="Use portal default")),
        ST(ENABLED, _("constraintypes_enable_label", default="Select manually")),
    ]
)


@implementer(IVocabularyFactory)
class ValidTypes:
    def __call__(self, context):
        constrain_aspect = context.context
        items = []
        for type_ in constrain_aspect.getDefaultAddableTypes():
            items.append(SimpleTerm(value=type_.getId(), title=type_.Title()))
        return SimpleVocabulary(items)


ValidTypesFactory = ValidTypes()


class IConstrainForm(Interface):
    constrain_types_mode = Choice(
        title=_("label_type_restrictions", default="Type restrictions"),
        description=_(
            "help_add_restriction_mode",
            default="Select the restriction policy " "in this location",
        ),
        vocabulary=possible_constrain_types,
        required=False,
    )

    allowed_types = List(
        title=_("label_immediately_addable_types", default="Allowed types"),
        description=_(
            "help_immediately_addable_types",
            default="Controls what types are addable " "in this location",
        ),
        value_type=Choice(source="plone.app.content.ValidAddableTypes"),
        required=False,
    )

    secondary_types = List(
        title=_("label_locally_allowed_types", default="Secondary types"),
        description=_(
            "help_locally_allowed_types",
            default=""
            "Select which types should be available in the "
            "'More&hellip;' submenu <em>instead</em> of in the "
            "main pulldown. "
            "This is useful to indicate that these are not the "
            "preferred types "
            "in this location, but are allowed if you really "
            "need them.",
        ),
        value_type=Choice(source="plone.app.content.ValidAddableTypes"),
        required=False,
    )

    @invariant
    def legal_not_immediately_addable(data):
        missing = []
        for one_allowed in data.secondary_types:
            if one_allowed not in data.allowed_types:
                missing.append(one_allowed)
        if missing:
            raise Invalid(
                _(
                    "You cannot have a type as a secondary type without "
                    "having it allowed. You have selected ${types}.",
                    mapping=dict(types=", ".join(missing)),
                )
            )
        return True


@implementer(IConstrainForm)
class FormContentAdapter:
    def __init__(self, context):
        self.context = ISelectableConstrainTypes(context)

    @property
    def constrain_types_mode(self):
        return self.context.getConstrainTypesMode()

    @property
    def allowed_types(self):
        return self.context.getLocallyAllowedTypes()

    @property
    def secondary_types(self):
        immediately_addable = self.context.getImmediatelyAddableTypes()
        return [
            t
            for t in self.context.getLocallyAllowedTypes()
            if t not in immediately_addable
        ]


class ConstrainsFormView(AutoExtensibleForm, form.EditForm):
    schema = IConstrainForm
    label = _(
        "heading_set_content_type_restrictions",
        default="Restrict what types of content can be added",
    )
    template = ViewPageTemplateFile("constraintypes.pt")

    def getContent(self):
        return FormContentAdapter(self.context)

    def updateFields(self):
        super().updateFields()
        self.fields["allowed_types"].widgetFactory = CheckBoxFieldWidget
        self.fields["secondary_types"].widgetFactory = CheckBoxFieldWidget

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["allowed_types"].addClass("current_prefer_form")
        self.widgets["secondary_types"].addClass("current_allow_form")
        self.widgets["constrain_types_mode"].addClass("constrain_types_mode_form")

    def updateActions(self):
        super().updateActions()
        self.actions["save"].addClass("btn btn-primary")

    @button.buttonAndHandler(_("label_save", default="Save"), name="save")
    def handleSave(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        allowed_types = data["allowed_types"]
        immediately_addable = [
            t for t in allowed_types if t not in data["secondary_types"]
        ]

        aspect = ISelectableConstrainTypes(self.context)
        aspect.setConstrainTypesMode(data["constrain_types_mode"])
        aspect.setLocallyAllowedTypes(allowed_types)
        aspect.setImmediatelyAddableTypes(immediately_addable)
        contextURL = self.context.absolute_url()
        self.request.response.redirect(contextURL)

    @button.buttonAndHandler(_("label_cancel", default="Cancel"), name="cancel")
    def handleCancel(self, action):
        contextURL = self.context.absolute_url()
        self.request.response.redirect(contextURL)



def reindexOnModify(content, event):
    """When an object is modified, re-index it in the catalog"""
    if event.object is not content:
        return
    content.reindexObject(idxs=getattr(event, "descriptions", []))



from plone.base.interfaces import INameFromTitle as FutureINameFromTitle
from zope.interface import Attribute
from zope.interface import Interface


class INameFromTitle(FutureINameFromTitle):
    """An object that supports getting its name (id) from its title.

    This interface has been moved to plone.base.interfaces.
    This alias will be removed in Plone 7.0.
    We tried deprecating it like this:

        zope.deferredimport.deprecated(
            INameFromTitle="plone.base.interfaces:INameFromTitle",
        )

    Unfortunately this does not completely work: if your site has a content
    type with behavior `plone.app.content.interfaces.INameFromTitle` this would
    no longer work because the behavior is not found.
    If you use `plone.namefromtitle` then it works.

    So as long as we want to support the old spelling, we must keep the
    interface here, and also use this interface as the `provides` in the
    definition of the behavior in `plone.app.dexterity`.

    See https://github.com/plone/plone.app.dexterity/pull/379
    """


class IReindexOnModify(Interface):
    """Marker interface which makes sure an object gets reindexed when
    it's modified.
    """


class IStructureAction(Interface):
    order = Attribute("Order the action should be listed")

    def get_options():
        """
        Return a dict of action widget options.

        Options: {
            'title': 'Button title', // required
            'name': 'short name', // required
            'formTemplate': None, // If action requires form to submit additional options
            'icon': None, // icon name
            'button-type': 'danger', //
        }
        """



from plone.app.content.browser.vocabulary import VocabularyView
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.z3cform.interfaces import IPloneFormLayer
from plone.app.z3cform.tests.layer import PAZ3CForm_INTEGRATION_TESTING
from plone.autoform.interfaces import WIDGETS_KEY
from plone.autoform.interfaces import WRITE_PERMISSIONS_KEY
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.fti import DexterityFTI
from z3c.form.interfaces import IFieldWidget
from z3c.form.util import getSpecification
from z3c.form.widget import FieldWidget
from zope import schema
from zope.component import provideAdapter
from zope.component.globalregistry import base
from zope.globalrequest import setRequest
from zope.interface import Interface
from zope.publisher.browser import TestRequest

import json
import unittest


def add_mock_fti(portal):
    # Fake DX Type
    fti = DexterityFTI("dx_mock")
    portal.portal_types._setObject("dx_mock", fti)
    fti.klass = "plone.dexterity.content.Item"
    fti.schema = "plone.app.content.tests.test_permissions.IMockSchema"
    fti.filter_content_types = False
    fti.behaviors = ("plone.app.dexterity.behaviors.metadata.IBasic",)


def _custom_field_widget(field, request):
    from plone.app.z3cform.widget import AjaxSelectWidget

    widget = FieldWidget(field, AjaxSelectWidget(request))
    widget.vocabulary = "plone.app.vocabularies.PortalTypes"
    return widget


class IMockSchema(Interface):
    allowed_field = schema.Choice(vocabulary="plone.app.vocabularies.PortalTypes")
    disallowed_field = schema.Choice(vocabulary="plone.app.vocabularies.PortalTypes")
    default_field = schema.Choice(vocabulary="plone.app.vocabularies.PortalTypes")
    custom_widget_field = schema.TextLine()
    adapted_widget_field = schema.TextLine()


IMockSchema.setTaggedValue(
    WRITE_PERMISSIONS_KEY,
    {
        "allowed_field": "zope2.View",
        "disallowed_field": "zope2.ViewManagementScreens",
        "custom_widget_field": "zope2.View",
        "adapted_widget_field": "zope2.View",
    },
)
IMockSchema.setTaggedValue(
    WIDGETS_KEY,
    {
        "custom_widget_field": _custom_field_widget,
    },
)


def _enable_custom_widget(field):
    provideAdapter(
        _custom_field_widget,
        adapts=(getSpecification(field), IPloneFormLayer),
        provides=IFieldWidget,
    )


def _disable_custom_widget(field):
    base.unregisterAdapter(
        required=(
            getSpecification(field),
            IPloneFormLayer,
        ),
        provided=IFieldWidget,
    )


class DexterityVocabularyPermissionTests(unittest.TestCase):
    layer = PAZ3CForm_INTEGRATION_TESTING

    def setUp(self):
        self.request = TestRequest(environ={"HTTP_ACCEPT_LANGUAGE": "en"})
        setRequest(self.request)
        self.portal = self.layer["portal"]

        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        add_mock_fti(self.portal)
        self.portal.invokeFactory("dx_mock", "test_dx")

        self.portal.test_dx.manage_permission("View", ("Anonymous",), acquire=False)
        self.portal.test_dx.manage_permission(
            "View management screens", (), acquire=False
        )
        self.portal.test_dx.manage_permission(
            "Modify portal content",
            ("Editor", "Manager", "Site Adiminstrator"),
            acquire=False,
        )

    def test_vocabulary_field_allowed(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "allowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]),
            len(self.portal.portal_types.objectIds()),
        )

    def test_vocabulary_field_wrong_vocabulary_disallowed(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Fake",
                "field": "allowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

    def test_vocabulary_field_disallowed(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "disallowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

    def test_vocabulary_field_default_permission(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "default_field",
            }
        )
        # If the field is does not have a security declaration, the
        # default edit permission is tested (Modify portal content)
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

        setRoles(self.portal, TEST_USER_ID, ["Editor"])
        # Now access should be allowed, but the vocabulary does not exist
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]),
            len(self.portal.portal_types.objectIds()),
        )

    def test_vocabulary_field_default_permission_wrong_vocab(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.Fake",
                "field": "default_field",
            }
        )
        setRoles(self.portal, TEST_USER_ID, ["Editor"])
        # Now access should be allowed, but the vocabulary does not exist
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

    def test_vocabulary_missing_field(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "missing_field",
            }
        )
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        with self.assertRaises(AttributeError):
            view()

    def test_vocabulary_on_widget(self):
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "custom_widget_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]),
            len(self.portal.portal_types.objectIds()),
        )
        self.request.form["name"] = "plone.app.vocabularies.Fake"
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")

    def test_vocabulary_on_adapted_widget(self):
        _enable_custom_widget(IMockSchema["adapted_widget_field"])
        view = VocabularyView(self.portal.test_dx, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "adapted_widget_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]),
            len(self.portal.portal_types.objectIds()),
        )

        self.request.form["name"] = "plone.app.vocabularies.Fake"
        data = json.loads(view())
        self.assertEqual(data["error"], "Vocabulary lookup not allowed")
        _disable_custom_widget(IMockSchema["adapted_widget_field"])

    def test_vocabulary_field_allowed_from_add_view(self):
        add_view = DefaultAddView(
            self.portal, self.request, self.portal.portal_types["dx_mock"]
        )
        view = VocabularyView(add_view, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "allowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]),
            len(self.portal.portal_types.objectIds()),
        )

    def test_vocabulary_field_allowed_from_add_form(self):
        add_form = DefaultAddForm(self.portal, self.request)
        add_form.portal_type = "dx_mock"
        view = VocabularyView(add_form, self.request)
        self.request.form.update(
            {
                "name": "plone.app.vocabularies.PortalTypes",
                "field": "allowed_field",
            }
        )
        data = json.loads(view())
        self.assertEqual(
            len(data["results"]),
            len(self.portal.portal_types.objectIds()),
        )



from AccessControl import getSecurityManager
from Acquisition import aq_base
from html import unescape
from logging import getLogger
from plone.app.content.utils import json_dumps
from plone.app.content.utils import json_loads
from plone.app.querystring import queryparser
from plone.app.z3cform.interfaces import IFieldPermissionChecker
from plone.autoform.interfaces import WRITE_PERMISSIONS_KEY
from plone.base import PloneMessageFactory as _
from plone.base.interfaces.siteroot import INavigationRoot
from plone.base.navigationroot import get_navigation_root
from plone.base.utils import safe_text
from plone.memoize.view import memoize
from plone.supermodel.utils import mergedTaggedValueDict
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.MimetypesRegistry.MimeTypeItem import guess_icon_path
from Products.MimetypesRegistry.MimeTypeItem import PREFIX
from Products.PortalTransforms.transforms.safe_html import SafeHTML
from types import FunctionType
from z3c.form.interfaces import IAddForm
from z3c.form.interfaces import ISubForm
from zope.component import getUtility
from zope.component import queryAdapter
from zope.component import queryUtility
from zope.deprecation import deprecated
from zope.i18n import translate
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import IVocabularyFactory
from zope.security.interfaces import IPermission

import inspect
import itertools


logger = getLogger(__name__)

MAX_BATCH_SIZE = 500  # prevent overloading server

DEFAULT_PERMISSION = "View"
DEFAULT_PERMISSION_SECURE = "Modify portal content"
PERMISSIONS = {
    "plone.app.vocabularies.Catalog": "View",
    "plone.app.vocabularies.Keywords": "Modify portal content",
    "plone.app.vocabularies.SyndicatableFeedItems": "Modify portal content",
    "plone.app.vocabularies.Users": "Modify portal content",
    "plone.app.multilingual.RootCatalog": "View",
}
TRANSLATED_IGNORED = [
    "author_name",
    "cmf_uid",
    "commentators",
    "created",
    "CreationDate",
    "Creator",
    "Date",
    "Description",
    "effective",
    "EffectiveDate",
    "end",
    "exclude_from_nav",
    "ExpirationDate",
    "expires",
    "getIcon",
    "getMimeIcon",
    "getId",
    "getObjSize",
    "getRemoteUrl",
    "getURL",
    "id",
    "in_response_to",
    "is_folderish",
    "last_comment_date",
    "listCreators",
    "location",
    "meta_type",
    "ModificationDate",
    "modified",
    "path",
    "portal_type",
    "review_state",
    "start",
    "Subject",
    "sync_uid",
    "Title",
    "total_comments" "UID",
]

_permissions = PERMISSIONS
deprecated("_permissions", "Use PERMISSIONS variable instead.")


def _parseJSON(s):
    # XXX this should be changed to a try loads except return s
    if isinstance(s, str):
        s = s.strip()
        if (s.startswith("{") and s.endswith("}")) or (
            s.startswith("[") and s.endswith("]")
        ):  # detect if json
            return json_loads(s)
    return s


_unsafe_metadata = [
    "author_name",
    "commentors",
    "Creator",
    "listCreators",
]
_safe_callable_metadata = [
    "getIcon",
    "getPath",
    "getURL",
    "is_folderish",
    "review_state",
]


class VocabLookupException(Exception):
    pass


class BaseVocabularyView(BrowserView):
    def get_translated_ignored(self):
        return TRANSLATED_IGNORED

    def get_base_path(self, context):
        return get_navigation_root(context)

    def __call__(self):
        """
        Accepts GET parameters of:
        name: Name of the vocabulary
        field: Name of the field the vocabulary is being retrieved for
        query: string or json object of criteria and options.
            json value consists of a structure:
                {
                    criteria: object,
                    sort_on: index,
                    sort_order: (asc|reversed)
                }
        attributes: comma separated, or json object list
        batch: {
            page: 1-based page of results,
            size: size of paged results
        }
        """
        context = self.get_context()
        self.request.response.setHeader(
            "Content-Type", "application/json; charset=utf-8"
        )

        try:
            vocabulary = self.get_vocabulary()
        except VocabLookupException as e:
            return json_dumps({"error": e.args[0]})

        results_are_brains = False
        if hasattr(vocabulary, "search_catalog"):
            query = self.parsed_query()
            results = vocabulary.search_catalog(query)
            results_are_brains = True
        elif hasattr(vocabulary, "search"):
            try:
                query = self.parsed_query()["SearchableText"]["query"]
            except KeyError:
                results = iter(vocabulary)
            else:
                results = vocabulary.search(query)
        else:
            results = vocabulary

        try:
            total = len(results)
        except TypeError:
            # do not error if object does not support __len__
            # we'll check again later if we can figure some size
            # out
            total = 0

        # get batch
        batch = _parseJSON(self.request.get("batch", ""))
        if batch and ("size" not in batch or "page" not in batch):
            batch = None  # batching not providing correct options
        if batch:
            # must be sliceable for batching support
            page = int(batch["page"])
            size = int(batch["size"])
            if size > MAX_BATCH_SIZE:
                raise Exception("Max batch size is 500")
            # page is being passed in is 1-based
            start = (max(page - 1, 0)) * size
            end = start + size
            # Try __getitem__-based slice, then iterator slice.
            # The iterator slice has to consume the iterator through
            # to the desired slice, but that shouldn't be the end
            # of the world because at some point the user will hopefully
            # give up scrolling and search instead.
            try:
                results = results[start:end]
            except TypeError:
                results = itertools.islice(results, start, end)

        # build result items
        items = []

        attributes = _parseJSON(self.request.get("attributes", ""))
        if isinstance(attributes, str) and attributes:
            attributes = attributes.split(",")

        translate_ignored = self.get_translated_ignored()
        transform = SafeHTML()
        if attributes:
            base_path = self.get_base_path(context)
            sm = getSecurityManager()
            can_edit = sm.checkPermission(DEFAULT_PERMISSION_SECURE, context)
            mtt = getToolByName(self.context, "mimetypes_registry")
            for vocab_item in results:
                if not results_are_brains:
                    vocab_item = vocab_item.value
                item = {}
                for attr in attributes:
                    key = attr
                    if ":" in attr:
                        key, attr = attr.split(":", 1)
                    if attr in _unsafe_metadata and not can_edit:
                        continue
                    if key == "path":
                        attr = "getPath"
                    val = getattr(vocab_item, attr, None)
                    if callable(val):
                        if attr in _safe_callable_metadata:
                            val = val()
                        else:
                            continue
                    if key == "path" and val is not None:
                        val = val[len(base_path) :]
                    if key not in translate_ignored and isinstance(val, str):
                        val = translate(_(safe_text(val)), context=self.request)
                    item[key] = val
                    if key == "getMimeIcon":
                        item[key] = None
                        # get mime type icon url from mimetype registry'
                        contenttype = aq_base(getattr(vocab_item, "mime_type", None))
                        if contenttype:
                            ctype = mtt.lookup(contenttype)
                            if ctype:
                                item[key] = "/".join(
                                    [base_path, guess_icon_path(ctype[0])]
                                )
                            else:
                                item[key] = "/".join(
                                    [
                                        base_path,
                                        PREFIX.rstrip("/"),
                                        "unknown.png",
                                    ]
                                )
                items.append(item)
        else:
            items = [
                {
                    "id": unescape(transform.scrub_html(item.value)),
                    "text": (
                        unescape(transform.scrub_html(item.title)) if item.title else ""
                    ),
                }
                for item in results
            ]

        if total == 0:
            total = len(items)

        return json_dumps({"results": items, "total": total})

    def parsed_query(
        self,
    ):
        query = _parseJSON(self.request.get("query", ""))
        if isinstance(query, str):
            query = {"SearchableText": {"query": query}}
        elif query:
            parsed = queryparser.parseFormquery(self.get_context(), query["criteria"])
            if "sort_on" in query:
                parsed["sort_on"] = query["sort_on"]
            if "sort_order" in query:
                parsed["sort_order"] = str(query["sort_order"])
            query = parsed
        else:
            query = {}
        return query


class VocabularyView(BaseVocabularyView):
    """Queries a named vocabulary and returns JSON-formatted results."""

    def get_context(self):
        return self.context

    def get_vocabulary(self):
        # Look up named vocabulary and check permission.

        context = self.context
        factory_name = self.request.get("name", None)
        field_name = self.request.get("field", None)
        if not factory_name:
            raise VocabLookupException("No factory provided.")
        authorized = None
        sm = getSecurityManager()
        if factory_name not in PERMISSIONS or not INavigationRoot.providedBy(context):
            # Check field specific permission
            if field_name:
                permission_checker = queryAdapter(context, IFieldPermissionChecker)
                if permission_checker is not None:
                    authorized = permission_checker.validate(field_name, factory_name)
                elif sm.checkPermission(
                    PERMISSIONS.get(factory_name, DEFAULT_PERMISSION), context
                ):
                    # If no checker, fall back to checking the global registry
                    authorized = True

            if not authorized:
                raise VocabLookupException("Vocabulary lookup not allowed")

        # Short circuit if we are on the site root and permission is
        # in global registry
        elif not sm.checkPermission(
            PERMISSIONS.get(factory_name, DEFAULT_PERMISSION), context
        ):
            raise VocabLookupException("Vocabulary lookup not allowed")

        factory = queryUtility(IVocabularyFactory, factory_name)
        if not factory:
            raise VocabLookupException(
                'No factory with name "%s" exists.' % factory_name
            )

        # This part is for backwards-compatibility with the first
        # generation of vocabularies created for plone.app.widgets,
        # which take the (unparsed) query as a parameter of the vocab
        # factory rather than as a separate search method.
        if isinstance(factory, FunctionType):
            factory_spec = inspect.getfullargspec(factory)
        else:
            factory_spec = inspect.getfullargspec(factory.__call__)
        query = _parseJSON(self.request.get("query", ""))
        if query and "query" in factory_spec.args:
            vocabulary = factory(context, query=query)
        else:
            # This is what is reached for non-legacy vocabularies.
            vocabulary = factory(context)

        return vocabulary


class SourceView(BaseVocabularyView):
    """Queries a field's source and returns JSON-formatted results."""

    def get_context(self):
        if ISubForm.providedBy(self.context.form):
            context = self.context.form.parentForm.context
        else:
            context = self.context.context
        return context

    @property
    @memoize
    def default_permission(self):
        if IAddForm.providedBy(self.context.form):
            return "cmf.AddPortalContent"
        return "cmf.ModifyPortalContent"

    def get_vocabulary(self):
        widget = self.context
        field = widget.field.bind(widget.context)

        # check field's write permission
        info = mergedTaggedValueDict(field.interface, WRITE_PERMISSIONS_KEY)
        permission_name = info.get(field.__name__, self.default_permission)
        permission = queryUtility(IPermission, name=permission_name)
        if permission is None:
            permission = getUtility(IPermission, name=self.default_permission)
        if not getSecurityManager().checkPermission(
            permission.title, self.get_context()
        ):
            raise VocabLookupException("Vocabulary lookup not allowed.")

        if ICollection.providedBy(field):
            return field.value_type.vocabulary
        return field.vocabulary



from plone.app.content.testing import PLONE_APP_CONTENT_DX_FUNCTIONAL_TESTING
from plone.app.content.testing import PLONE_APP_CONTENT_DX_INTEGRATION_TESTING
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.base.utils import is_expired
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedBlobImage
from plone.testing.zope import Browser
from zope.component import getMultiAdapter

import os
import transaction
import unittest


class TestContentStatusModifyIntegration(unittest.TestCase):
    """The the content_status_modify view.

    Until and including Plone 5.2, this was a skin script in Products.CMFPlone.
    Tests adapted from CMFPlone/tests/testContentPublishing.py.
    """

    layer = PLONE_APP_CONTENT_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.workflow = self.portal.portal_workflow
        # Make sure we can create and publish directly.
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        # Prepare content.
        self.portal.invokeFactory("Folder", id="folder")
        self.folder = self.portal.folder
        self.folder.invokeFactory("Document", id="d1", title="Doc 1")

    def setup_authenticator(self):
        from plone.protect.authenticator import createToken

        self.request.form["_authenticator"] = createToken()

    def get_content_status_modify_view(self, obj):
        self.setup_authenticator()
        view = getMultiAdapter((obj, self.request), name="content_status_modify")
        return view

    # Test the recursive behaviour of content_status_modify:

    def testPublishingNonDefaultPageLeavesFolderAlone(self):
        view = self.get_content_status_modify_view(self.folder.d1)
        view("publish")
        self.assertEqual(
            self.workflow.getInfoFor(self.folder, "review_state"), "private"
        )
        self.assertEqual(
            self.workflow.getInfoFor(self.folder.d1, "review_state"), "published"
        )

    def testPublishingDefaultPagePublishesFolder(self):
        self.folder.setDefaultPage("d1")
        view = self.get_content_status_modify_view(self.folder.d1)
        view("publish")
        self.assertEqual(
            self.workflow.getInfoFor(self.folder, "review_state"), "published"
        )
        self.assertEqual(
            self.workflow.getInfoFor(self.folder.d1, "review_state"), "published"
        )

    def testPublishingDefaultPageWhenFolderCannotBePublished(self):
        self.folder.setDefaultPage("d1")
        # make parent be published already when publishing its default document
        # results in an attempt to do it again
        view = self.get_content_status_modify_view(self.folder)
        view("publish")
        self.assertEqual(
            self.workflow.getInfoFor(self.folder, "review_state"), "published"
        )
        view = self.get_content_status_modify_view(self.folder.d1)
        view("publish")
        self.assertEqual(
            self.workflow.getInfoFor(self.folder, "review_state"), "published"
        )
        self.assertEqual(
            self.workflow.getInfoFor(self.folder.d1, "review_state"), "published"
        )

    # test setting effective/expiration date and isExpired method

    def testIsExpiredWithExplicitExpiredContent(self):
        view = self.get_content_status_modify_view(self.folder.d1)
        view(
            workflow_action="publish",
            effective_date="1/1/2001",
            expiration_date="1/2/2001",
        )
        self.assertTrue(is_expired(self.folder.d1))

    def testIsExpiredWithExplicitNonExpiredContent(self):
        view = self.get_content_status_modify_view(self.folder.d1)
        view(workflow_action="publish")
        self.assertFalse(is_expired(self.folder.d1))

    def testEditorCanSubmitButNotPublish(self):
        setRoles(self.portal, TEST_USER_ID, ["Contributor"])
        self.folder.invokeFactory("Document", id="d2", title="Doc 2")
        view = self.get_content_status_modify_view(self.folder.d2)
        view(workflow_action="submit")
        self.assertEqual(
            self.workflow.getInfoFor(self.folder.d2, "review_state"), "pending"
        )

        # Now try publishing.
        # For various reasons, there are no complaints/errors when trying
        # a transition that you are not allowed to do.
        view = self.get_content_status_modify_view(self.folder.d2)
        view(workflow_action="publish")
        self.assertEqual(
            self.workflow.getInfoFor(self.folder.d2, "review_state"), "pending"
        )


class TestContentStatusModifyFunctional(unittest.TestCase):
    layer = PLONE_APP_CONTENT_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        login(self.portal, TEST_USER_NAME)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        # Set workflow for files and images.
        wf_tool = self.portal.portal_workflow
        wf_tool.setChainForPortalTypes(["File"], "simple_publication_workflow")
        wf_tool.setChainForPortalTypes(["Image"], "simple_publication_workflow")

        # Create content.
        filename = os.path.join(os.path.dirname(__file__), "image.png")
        with open(filename, "rb") as f:
            FILE_DATA = f.read()
        self.portal.invokeFactory("Document", id="d1", title="Doc 1")
        self.portal.invokeFactory(
            "File",
            id="f1",
            title="File 1",
            file=NamedBlobFile(data=FILE_DATA, filename="image.png"),
        )
        self.portal.invokeFactory(
            "Image",
            id="i1",
            title="Image 1",
            image=NamedBlobImage(data=FILE_DATA, filename="image.png"),
        )
        self.portal_url = self.portal.absolute_url()
        transaction.commit()

        # Setup the test browser.
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False
        self.browser.raiseHttpErrors = False
        self.browser.addHeader(
            "Authorization", f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}"
        )

    def test_history_doc(self):
        self.browser.open(f"{self.portal_url}/d1/content_status_history")
        self.browser.getControl("Save").click()
        self.assertEqual(self.browser.url, f"{self.portal_url}/d1")

    def test_history_file(self):
        self.browser.open(f"{self.portal_url}/f1/content_status_history")
        self.browser.getControl("Save").click()
        self.assertEqual(self.browser.url, f"{self.portal_url}/f1/view")

    def test_history_image(self):
        self.browser.open(f"{self.portal_url}/i1/content_status_history")
        self.browser.getControl("Save").click()
        self.assertEqual(self.browser.url, f"{self.portal_url}/i1/view")



from DateTime import DateTime
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.interfaces import IStructureAction
from plone.app.dexterity.behaviors.metadata import ICategorization
from plone.app.z3cform.widgets.datetime import get_date_options
from plone.base import PloneMessageFactory as _
from plone.base.defaultpage import check_default_page_via_view
from Products.CMFCore.interfaces._content import IFolderish
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.i18n import translate
from zope.interface import implementer
from zope.schema.interfaces import IVocabularyFactory

import json


@implementer(IStructureAction)
class PropertiesAction:
    template = ViewPageTemplateFile("templates/properties.pt")
    order = 8

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_options(self):
        base_vocabulary = "%s/@@getVocabulary?name=" % getSite().absolute_url()
        return {
            "tooltip": translate(_("Properties"), context=self.request),
            "id": "properties",
            "icon": "plone-edit",
            "url": self.context.absolute_url() + "/@@fc-properties",
            "form": {
                "title": translate(
                    _("Modify properties on items"), context=self.request
                ),
                "template": self.template(
                    vocabulary_url="%splone.app.vocabularies.Users" % (base_vocabulary),
                    pattern_options=json.dumps(get_date_options(self.request)),
                ),
                "dataUrl": self.context.absolute_url() + "/@@fc-properties",
            },
        }


class PropertiesActionView(ContentsBaseAction):
    success_msg = _("Successfully updated metadata")
    failure_msg = _("Failure updating metadata")
    required_obj_permission = "Modify portal content"

    def __call__(self):
        if self.request.form.get("render") == "yes":
            lang_factory = getUtility(
                IVocabularyFactory, "plone.app.vocabularies.SupportedContentLanguages"
            )
            lang_vocabulary = lang_factory(self.context)
            languages = [
                {"title": term.title, "value": term.value} for term in lang_vocabulary
            ]
            return self.json(
                {
                    "languages": [
                        {
                            "title": translate(
                                _("label_no_change", default="No change"),
                                context=self.request,
                            ),
                            "value": "",
                        }
                    ]
                    + languages
                }
            )

        self.effectiveDate = self.request.form.get("effectiveDate")
        self.expirationDate = self.request.form.get("expirationDate")
        self.copyright = self.request.form.get("copyright")
        self.contributors = self.request.form.get("contributors")
        if self.contributors:
            self.contributors = self.contributors.split(",")
        else:
            self.contributors = []
        self.creators = self.request.form.get("creators", "")
        if self.creators:
            self.creators = self.creators.split(",")
        self.exclude = self.request.form.get("exclude-from-nav")
        self.language = self.request.form.get("language")
        self.recurse = self.request.form.get("recurse", "no") == "yes"
        return super().__call__()

    def action(self, obj, bypass_recurse=False):
        if check_default_page_via_view(obj, self.request):
            self.action(obj.aq_parent, bypass_recurse=True)
        recurse = self.recurse and not bypass_recurse
        if recurse and IFolderish.providedBy(obj):
            for sub in obj.values():
                self.action(sub)

        if self.effectiveDate and hasattr(obj, "effective_date"):
            obj.effective_date = DateTime(self.effectiveDate)
        if self.expirationDate and hasattr(obj, "expiration_date"):
            obj.expiration_date = DateTime(self.expirationDate)
        if self.copyright and hasattr(obj, "rights"):
            obj.rights = self.copyright
        if self.contributors and hasattr(obj, "contributors"):
            obj.contributors = tuple(self.contributors)
        if self.creators and hasattr(obj, "creators"):
            obj.creators = tuple(self.creators)
        if self.exclude and hasattr(obj, "exclude_from_nav"):
            obj.exclude_from_nav = self.exclude == "yes"

        behavior_categorization = ICategorization(obj)
        if self.language and behavior_categorization:
            behavior_categorization.language = self.language

        obj.reindexObject()



from DateTime import DateTime
from plone.app.content.browser.contents import ContentsBaseAction
from plone.app.content.interfaces import IStructureAction
from plone.base import PloneMessageFactory as _
from plone.base.defaultpage import check_default_page_via_view
from plone.base.utils import safe_text
from Products.CMFCore.interfaces._content import IFolderish
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from ZODB.POSException import ConflictError
from zope.i18n import translate
from zope.interface import implementer


@implementer(IStructureAction)
class WorkflowAction:
    template = ViewPageTemplateFile("templates/workflow.pt")
    order = 7

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_options(self):
        return {
            "tooltip": translate(_("State"), context=self.request),
            "id": "workflow",
            "icon": "plone-lock",
            "url": self.context.absolute_url() + "/@@fc-workflow",
            "form": {
                "title": translate(
                    _("Change workflow of selected items"), context=self.request
                ),
                "template": self.template(),
                "dataUrl": self.context.absolute_url() + "/@@fc-workflow",
            },
        }


class WorkflowActionView(ContentsBaseAction):
    required_obj_permission = "Modify portal content"
    success_msg = _("Successfully modified items")
    failure_msg = _("Failed to modify items")

    def __call__(self):
        self.pworkflow = getToolByName(self.context, "portal_workflow")
        self.transition_id = self.request.form.get("transition", None)
        self.comments = self.request.form.get("comments", "")
        self.recurse = self.request.form.get("recurse", "no") == "yes"
        if self.request.form.get("render") == "yes":
            # asking for render information
            selection = self.get_selection()
            catalog = getToolByName(self.context, "portal_catalog")
            brains = catalog(UID=selection, show_inactive=True)
            transitions = []
            for brain in brains:
                obj = brain.getObject()
                for transition in self.pworkflow.getTransitionsFor(obj):
                    tdata = {
                        "id": transition["id"],
                        "title": self.context.translate(safe_text(transition["name"])),
                    }
                    if tdata not in transitions:
                        transitions.append(tdata)
            return self.json({"transitions": transitions})
        return super().__call__()

    def action(self, obj, bypass_recurse=False):
        transitions = self.pworkflow.getTransitionsFor(obj)
        if self.transition_id in [t["id"] for t in transitions]:
            try:
                # set effective date if not already set
                if obj.EffectiveDate() == "None":
                    obj.setEffectiveDate(DateTime())

                self.pworkflow.doActionFor(
                    obj, self.transition_id, comment=self.comments
                )
                if check_default_page_via_view(obj, self.request):
                    self.action(obj.aq_parent, bypass_recurse=True)
                recurse = self.recurse and not bypass_recurse
                if recurse and IFolderish.providedBy(obj):
                    for sub in obj.values():
                        self.action(sub)
                obj.reindexObject()
            except ConflictError:
                raise
            except Exception:
                self.errors.append(
                    _(
                        "Could not transition: ${title}",
                        mapping={"title": self.objectTitle(obj)},
                    )
                )
